#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import threading, time


class StoppableThread(threading.Thread):

    def __init__(self,  *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()


class Spinner():

    def __init__(self, target = None, rate = 1.0):
        self._delay  = 1. / abs(rate)
        self._target = target
        self._thread = None
        self.real_rate = 0.0


    def _rated_loop(self, *args, **kwargs):
        current_thread = threading.currentThread()
        next_call_time = time.time()
        while not current_thread.stopped():
            next_call_time = next_call_time + self._delay;
            self._target(*args, **kwargs)
            rest_time = next_call_time - time.time()
            self.real_rate = (self.real_rate + min(1. / rest_time, 0))
            time.sleep(max(rest_time, 0))


    def set_target(self, target):
        self.stop()
        self._target = target


    def set_rate(self, rate):
        self._delay  = 1. / abs(rate)


    def is_active(self):
        return (self._thread and self._thread.is_alive())


    def is_not_active(self):
        return not self.is_active()


    def start(self,  *args, **kwargs):
        self.stop()
        self._thread = StoppableThread(target = self._rated_loop, args=args, kwargs=kwargs)
        self._thread.daemon = True
        self._thread.start()


    def stop(self):
        if self._thread and self._thread.is_alive():
            self._thread.stop()


    def restart(self):
        self.stop()
        self.start()


    def get_real_rate(self):
        return self.real_rate



#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import sys
import io
import log

from serial import *

from loops import Spinner
from slp import *


SERIAL_WRITE_RATE = 100.
SERIAL_READ_RATE  = 100.


def create_interface_by_name(interface, protocol):
    '''
    Takes interface name or path and OscarProtocol object as parameters and
    return corresponding OscarVehicleInterface object.
    '''
    if (interface == '/dev/ttyACM0'):
        try:
            return OscarSerialVehicleInterface(interface, protocol)
        except Exception as e:
            log.print_err("[OSCAR]: " + str(e))
            return None

    elif(interface == '' or interface is None):
        log.print_warn("[OSCAR] : Communication interface was not specified!")
        return None

    else:
        log.print_err("[OSCAR]: Communication interface type of " + str(interface) + " is not supported!")
        return None


class OscarVehicleInterface():

    def __init__(self, protocol):

        self._protocol = protocol

    def start_communication(self):
        return (self.start_sender() and self.start_receiver())


    def stop_communication(self):
        return (self.stop_sender() and self.stop_receiver())


    def start_sender(self):
        return False


    def stop_sender(self):
        return False


    def start_receiver(self):
        return False


    def stop_receiver(self):
        return False


class OscarSerialVehicleInterface(OscarVehicleInterface):
    '''
    StarLineProtocol (SLP) over Serial communication interface.
    Directly ask OscarProtocol for what data is needed to send and
    to update received data.
    '''
    def __init__(self, port, protocol):

        OscarVehicleInterface.__init__(self, protocol)
        self._serial_read_buffer = bytearray()
        self._serial = Serial(port=port, baudrate=115200, bytesize=EIGHTBITS)

        self._receive_spinner = Spinner(self._receive_by_protocol, SERIAL_READ_RATE)
        self._send_spinner    = Spinner(self._send_by_protocol, SERIAL_WRITE_RATE)


    def _send_by_protocol(self):
        # TODO: max size data constraints (serial buf on receiver side has limited size)
        oscar_can_raw_data_array = self._protocol.get_raw_data_to_send()
        for raw_data in oscar_can_raw_data_array:
            raw_slp = create_raw_slp_from_raw_oscar_data(raw_data)
            self._serial.write(raw_slp)


    def _receive_by_protocol(self):

        self._serial_read_buffer += self._serial.read(self._serial.in_waiting)
        read_time = time.time()
        buffer_len = len(self._serial_read_buffer)
        end_buffer_check_pose = buffer_len - RAW_SLP_LEN
        oscar_can_raw_data_array = []
        left_buffer_pose = 0

        i = 0
        while (i <= end_buffer_check_pose):

            slp_data = get_slp_data_from_raw_slp(self._serial_read_buffer[i:i+RAW_SLP_LEN])

            if slp_data:
                i += RAW_SLP_LEN
                left_buffer_pose = i
                oscar_can_raw_data_array.append(slp_data)
            else:
                i += 1


        if oscar_can_raw_data_array:
            self._protocol.update_data_from_raw(oscar_can_raw_data_array, read_time)
            self._serial_read_buffer = self._serial_read_buffer[left_buffer_pose:]


    def start_communication(self):
        return (self.start_sender() and self.start_receiver())


    def stop_communication(self):
        return (self.stop_sender() and self.stop_receiver())


    def start_sender(self):
        self._send_spinner.start()
        return self._send_spinner.is_active()


    def stop_sender(self):
        self._send_spinner.stop()
        return not self._send_spinner.is_active()


    def start_receiver(self):
        self._receive_spinner.start()
        return self._receive_spinner.is_active()


    def stop_receiver(self):
        self._receive_spinner.stop()
        return not self._receive_spinner.is_active()


    def __del__(self):
        if hasattr(self, 'serial'):
            self._serial.flush()
            self._serial.close()



#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import os, sys, time

import json, uuid
from collections import OrderedDict as odict


def error(msg):
    sys.exit(msg)


def print_ok(msg):
    print('\x1b[1;32m' + msg + '\x1b[0m')


def print_err(msg):
    print('\x1b[1;31;40m' + msg + '\x1b[0m')


def print_warn(msg):
    print('\x1b[1;33;40m' + msg + '\x1b[0m')


class NoIndent(object):
    def __init__(self, value):
        self.value = value


class NoIndentEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super(NoIndentEncoder, self).__init__(*args, **kwargs)
        self.kwargs = dict(kwargs)
        del self.kwargs['indent']
        self._replacement_map = {}

    def default(self, o):
        if isinstance(o, NoIndent):
            key = uuid.uuid4().hex
            self._replacement_map[key] = json.dumps(o.value, **self.kwargs)
            return "@@%s@@" % (key,)
        else:
            return super(NoIndentEncoder, self).default(o)

    def encode(self, o):
        result = super(NoIndentEncoder, self).encode(o)
        for k, v in self._replacement_map.iteritems():
            result = result.replace('"@@%s@@"' % (k,), v)
        return result


class VehicleLog():

    def __init__(self, log_dir = "/tmp"):
        self._zero_time = time.time()
        self._id_counter = 0
        self._log_dir = log_dir
        self._data = odict([
            ('header', odict([('ctime', time.strftime("%m.%d.%Y_%H.%M.%S", time.localtime())),
                              ('duration', 0.0),
                              ('size',   0)])),
            ('data', [])
        ])


    def add_data(self, dtime, sw_angle = 0.0,
                              sw_velocity = 0.0,
                              sw_torque = 0.0,
                              eps_torque = 0.0,
                              vehicle_speed = 0.0,
                              mode = 0):

        dtime -= self._zero_time

        self._data['data'].append(NoIndent(odict([('id', self._id_counter),
                                                  ('t', format(dtime, '.4f')),
                                                  ('sw_angle', format(sw_angle, '.1f')),
                                                  ('sw_velocity', format(sw_velocity, '.2f')),
                                                  ('sw_torque', format(sw_torque, '.1f')),
                                                  ('eps_torque', format(eps_torque, '.1f')),
                                                  ('vehicle_speed', format(vehicle_speed, '.2f')),
                                                  ('mode', mode)])))

        self._id_counter += 1


    def _open_file_to_write(self, file_path):

        try:
            file_handler = open(file_path, 'w')
        except Exception as err:
            log.print_err("[OSCAR]: Can't open file " + str(file_path))
            log.print_err(err)
            return None

        return file_handler


    def save(self):

        if not os.path.exists(self._log_dir):
            os.makedirs(self._log_dir)

        file_path = self._log_dir + "/oscar_vehicle_" + self._data['header']['ctime'] + ".log"

        duration = float(self._data['data'][-1].value['t']) - float(self._data['data'][0].value['t'])
        self._data['header']['duration'] = format(duration, '.4f')

        self._data['header']['size'] = len(self._data['data'])
        file_handler = self._open_file_to_write(file_path)

        if file_handler:
            file_handler.write(json.dumps(self._data, indent=2, cls=NoIndentEncoder))
            file_handler.flush()
            file_handler.close()
            return True
        else:
            return False



###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

from math import copysign
import threading
import time

import log

OSCAR_CAN_V2 = 2
OSCAR_CAN_V3 = 3

UNKNOWN                   = 0x00
VEHICLE_AUTO_MODE         = 0x01
VEHICLE_MANUAL_MODE       = 0x02

# Commands CAN identifiers
STEERING_WHEEL_TORQUE_CMD_V2 = 0x7F9
VEHICLE_ACCELERATION_CMD_V2  = 0x7EE
LAUNCHER_CMD_V2              = 0x7FC
EMERGENCY_STOP_CMD_V2        = 0x7FC
HAND_BRAKE_CMD_V2            = 0x7FC
INFO_CONFIGURATION_CMD_V2    = 0x7FD
TURN_SIGNALS_CMD_V2          = 0x778

# Infos CAN identifiers
STEERING_WHEEL_POSE_VELOCITY_INFO_V2  = 0x25
STEERING_WHEEL_EPS_TORQUE_INFO_V2     = 0x260
STEERING_WHEEL_TORQUE_INFO_V2         = 0x7F1
STEERING_WHEEL_INTERCEPTION_INFO_V2   = 0x7F1
VEHICLE_SPEED_INFO_V2                 = 0xB4
VEHICLE_WHEELS_SPEED_INFO_V2          = 0xAA
VEHICLE_MOVING_INTERCEPTION_INFO_V2   = 0x7F5
LAUNCHER_INFO_V2                      = 0x7F6
EMERGENCY_STOP_INFO_V2                = 0x7F6
HAND_BRAKE_INFO_V2                    = 0x7F6


class OscarProtocolConfig():
    """
    Functionality-oriented protocol configuration structure.
    """
    def __init__(self, version = OSCAR_CAN_V2):

        self.version = version

        self.vehicle_speed_info_need_to_receive                = True
        self.vehicle_wheels_speed_info_need_to_receive         = False
        self.vehicle_moving_interception_info_need_to_receive  = True

        self.steering_wheel_pose_velocity_info_need_to_receive = True
        self.steering_wheel_eps_torques_info_need_to_receive   = True
        self.steering_wheel_torque_info_need_to_receive        = True
        self.steering_wheel_interception_info_need_to_receive  = True

        self.launcher_info_need_to_receive                     = True
        self.emergency_stop_info_need_to_receive               = True
        self.hand_brake_info_need_to_receive                   = True

        self.steering_wheel_torque_cmd_send_rate               = 60.0
        self.vehicle_move_cmd_send_rate                        = 60.0
        self.emergency_stop_cmd_rate                           = 30.0

        self.launcher_info_rate                     = 2.0
        self.emergency_stop_info_rate               = 2.0
        self.hand_brake_info_rate                   = 2.0

        self.vehicle_moving_interception_info_rate  = 2.0
        self.steerin_wheel_interception_info_rate   = 2.0

        self.vehicle_speed_info_rate                = 80.0

        self.steering_wheel_pose_velocity_info_rate = 80.0
        self.steering_wheel_torque_info_rate        = 80.0


class OscarProtocol():

    SEND_ONCE = 1
    SEND_ONCE_NEED_REPLY = 2
    SEND_CONSTANTLY = 3
    SEND_CONSTANTLY_NEED_REPLY = 4

    RECEIVE_ONCE = 1
    RECEIVE_CONSTANTLY = 2


    def __init__(self, config = None):

        # All commands handlers
        self.steering_wheel_torque_cmd = None
        self.vehicle_move_cmd  = None

        self.launcher_cmd              = None
        self.emergency_stop_cmd        = None
        self.hand_brake_cmd            = None

        self.turn_signals_cmd          = None

        self.info_configuration_cmd    = None

        # All infos handlers

        self.vehicle_speed_info                = None
        self.vehicle_wheels_speed_info         = None
        self.vehicle_moving_interception_info  = None

        self.steering_wheel_pose_velocity_info = None
        self.steering_wheel_eps_torques_info   = None
        self.steering_wheel_torque_info        = None
        self.steering_wheel_interception_info  = None

        self.launcher_info                     = None
        self.emergency_stop_info               = None
        self.hand_brake_info                   = None

        self._send_list_lock    = threading.Lock()
        self._receive_list_lock = threading.Lock()

        self._send_list    = {}
        self._receive_list = {}

        self._identifier_counter = 0

        self._configure(config)


    def _configure(self, config):

        if not config:
            config = OscarProtocolConfig()
            log.print_warn("Default OSCAR Protocol configuration (OSCAR_CAN_V2)")

        if config.version == OSCAR_CAN_V2:

            # All commands handlers
            self.steering_wheel_torque_cmd       = OscarSteeringWheelTorqueCmdDataV2(self)
            self.steering_wheel_interception_cmd = OscarSteeringWheelTorqueCmdDataV2(self)

            self.vehicle_move_cmd                = OscarVehicleMoveCmdDataV2(self)
            self.vehicle_move_interception_cmd   = OscarVehicleMoveCmdDataV2(self)

            self.info_configuration_cmd          = OscarInfoConfigurationCmdDataV2(self)

            self.launcher_cmd                    = OscarLauncherCmdDataV2(self)
            self.emergency_stop_cmd              = OscarLauncherCmdDataV2(self)
            self.hand_brake_cmd                  = OscarLauncherCmdDataV2(self)

            self.turn_signals_cmd                = OscarTurnSignalsCmdDataV2(self)

            # All infos handlers
            self.vehicle_speed_info                = OscarVehicleSpeedInfoDataV2(self)
            self.vehicle_wheels_speed_info         = OscarVehicleWheelsSpeedInfoDataV2(self)
            # self.vehicle_moving_interception_info  = OscarDsuLkaInterceptionInfoDataV2(self)

            self.steering_wheel_pose_velocity_info = OscarSteeringWheelPoseVelocityInfoDataV2(self)
            self.steering_wheel_eps_torques_info   = OscarSteeringWheelEpsTorquesInfoDataV2(self)
            # self.steering_wheel_torque_info        = OscarSteeringWheelTorqueInfoDataV2(self)
            # self.steering_wheel_interception_info  = self.steering_wheel_torque_info

            self.launcher_info                     = OscarLauncherInfoDataV2(self)
            self.emergency_stop_info               = self.launcher_info
            self.hand_brake_info                   = self.launcher_info

        # All commands setting up
        self.steering_wheel_torque_cmd.set_send_rate(config.steering_wheel_torque_cmd_send_rate)
        self.vehicle_move_cmd.set_send_rate(config.vehicle_move_cmd_send_rate)
        self.emergency_stop_cmd.set_send_rate(config.emergency_stop_cmd_rate)

        # All infos setting up
        if config.vehicle_speed_info_need_to_receive:
            self.vehicle_speed_info.start_receiving()

        if config.vehicle_wheels_speed_info_need_to_receive:
            self.vehicle_wheels_speed_info.start_receiving()

        if config.steering_wheel_pose_velocity_info_need_to_receive:
            self.steering_wheel_pose_velocity_info.start_receiving()

        if config.steering_wheel_eps_torques_info_need_to_receive:
            self.steering_wheel_eps_torques_info.start_receiving()

        if config.launcher_info_need_to_receive:    # emergency_stop and hand_brake
            self.launcher_info.start_receiving()
            self.launcher_info.set_receive_rate(config.launcher_info_rate)


    def _get_identifier(self):
        identifier = self._identifier_counter
        self._identifier_counter += 1
        return identifier


    def _copy_data(self, data):
        new_data = globals()[data.__class__.__name__](self)
        new_data._can_data  = data._can_data
        new_data._send_type = data._send_type
        return new_data


    def _register_to_send_list(self, data):
        with self._send_list_lock:
            if data._send_type == self.SEND_ONCE:
                data = self._copy_data(data)
                data._identifier = self._get_identifier()
            self._send_list.update({data._identifier: data})


    def _get_actual_need_to_send_data(self):
        current_time = time.time()
        need_to_send_list = {}
        with self._send_list_lock:
            for data in self._send_list.itervalues():
                data_time_left = data._next_send_time - current_time

                if (data_time_left <= 0.0):
                    need_to_send_list.update({data._identifier: data})

        return need_to_send_list


    def _unregister_from_send_list(self, data):
        with self._send_list_lock:
            self._send_list.pop(data._identifier, None)


    def _register_to_receive_list(self, data):
        self._receive_list.update({data._get_can_id(): data})


    def _unregister_from_receive_list(self, data):
        self._receive_list.pop(data._get_can_id(), None)


    def _configure_received_infos_rate(self, data, rate):
        if self.info_configuration_cmd:
            if rate > 0:
                self.info_configuration_cmd.set_reply_delay(data, 1 / rate)
            else:
                self.info_configuration_cmd.turn_off_replies(data)
            return True
        else:
            return False


    def get_raw_data_to_send(self):
        """
        Return array of protocol raw data which have to be sent by interface.
        """
        raw_data_frames_to_send = []
        need_to_send_data_list = self._get_actual_need_to_send_data()

        if need_to_send_data_list:
            current_time = time.time()
            for data in need_to_send_data_list.itervalues():
                data_time_left = data._next_send_time - current_time
                raw_data_frames_to_send.append(data.raw())

                if data._send_type == self.SEND_CONSTANTLY:
                    data._next_send_time = current_time + 1./data._send_rate
                    data._real_send_rate = data._send_rate + data_time_left

                elif data._send_type == self.SEND_ONCE:
                    self._unregister_from_send_list(data)

        return raw_data_frames_to_send


    def update_data_from_raw(self, raws, receive_time = 0.0):

        for raw in raws:
            can_id = raw[1] + (raw[0] << 8)
            if can_id in self._receive_list:
                self._receive_list[can_id].update_from_raw(raw, receive_time)


    def auto_mode(self):
        self.launcher_cmd.auto_mode()
        return True


    def manual_mode(self):
        self.launcher_cmd.manual_mode()
        return True


    def get_mode(self):
        return self.launcher_info.get_mode()


    def emergency_stop_on(self):
        self.emergency_stop_cmd.emergency_stop_on()
        return True


    def emergency_stop_off(self):
        self.emergency_stop_cmd.emergency_stop_off()
        return True


    def get_emergency_stop(self):
        emergency_stop_state, source = self.launcher_info.get_emergency_stop()
        return (emergency_stop_state == self.launcher_info.EMERGENCY_STOP_ON)


    def hand_brake_on(self):
        self.hand_brake_cmd.hand_brake_on()
        return True


    def hand_brake_off(self):
        self.hand_brake_cmd.hand_brake_off()
        return True


    def get_hand_brake(self):
        hand_brake_state, source = self.launcher_info.get_hand_brake()
        return (hand_brake_state == self.launcher_info.HAND_BRAKE_ON)


    def led_on(self):
        self.launcher_cmd.led_on()
        return True


    def led_off(self):
        self.launcher_cmd.led_off()
        return True


    def get_led(self):
        return (self.launcher_info.get_led() == self.launcher_info.LED_ON)


    def led_reverse(self):
        if (self.launcher_info.get_led() == self.launcher_info.LED_ON):
            self.launcher_cmd.led_off()
        else:
            self.launcher_cmd.led_on()


    def left_turn_signal(self):
        self.turn_signals_cmd.left_signal()


    def right_turn_signal(self):
        self.turn_signals_cmd.right_signal()


    def emergency_signals(self):
        self.turn_signals_cmd.emergency_signals()


    def turn_off_signals(self):
        self.turn_signals_cmd.turn_off_signals()


    def vehicle_move_interception_on(self):
        self.vehicle_move_interception_cmd.interception_on()
        return True


    def vehicle_move_interception_off(self):
        self.vehicle_move_interception_cmd.interception_off()
        return True


    def start_sending_vehicle_move_cmd(self):
        self.vehicle_move_cmd.start_sending()
        return True


    def stop_sending_vehicle_move_cmd(self):
        self.vehicle_move_cmd.stop_sending()
        return True


    def set_vehicle_throttle(self, throttle):
        self.vehicle_move_cmd.throttle(throttle)


    def get_vehicle_speed(self):
        return self.vehicle_speed_info.get_vehicle_speed()


    def get_vehicle_wheels_speed(self):
        return self.vehicle_wheels_speed_info.get_vehicle_wheels_speed()


    def steering_wheel_interception_on(self):
        self.steering_wheel_interception_cmd.interception_on()
        return True


    def steering_wheel_interception_off(self):
        self.steering_wheel_interception_cmd.interception_off()
        return True


    def start_sending_steering_wheel_torque_cmd(self):
        self.steering_wheel_torque_cmd.start_sending()
        return True


    def stop_sending_steering_wheel_torque_cmd(self):
        self.steering_wheel_torque_cmd.stop_sending()
        return True


    def set_steering_wheel_torque(self, steering_wheel_torque):
        self.steering_wheel_torque_cmd.torque(steering_wheel_torque)


    def get_steering_wheel_angle_and_velocity(self):
        return self.steering_wheel_pose_velocity_info.get_steering_wheel_angle_and_velocity()


    def get_steering_wheel_and_eps_torques(self):
        return self.steering_wheel_eps_torques_info.get_steering_wheel_and_eps_torques()



class OscarData(object):

    CAN_TYPE_INFO = bytearray.fromhex("01")
    CAN_TYPE_CMD  = bytearray.fromhex("02")
    CAN_TYPE_ACK  = bytearray.fromhex("03")

    CAN_STATE_UNKNOWN             = bytearray.fromhex("00")
    CAN_STATE_WORK_AND_ACTIVE     = bytearray.fromhex("01")
    CAN_STATE_WORK_BUT_NOT_ACTIVE = bytearray.fromhex("02")
    CAN_STATE_ERROR               = bytearray.fromhex("03")
    CAN_STATE_DEBUG               = bytearray.fromhex("04")

    def __init__(self, protocol):
        self._protocol = protocol
        self._identifier = self._protocol._get_identifier()

        self._send_rate = -1
        self._real_send_rate = 0.0
        self._next_send_time = 0.0
        self._send_type = None

        self._receive_rate = None
        self._real_receive_rate = 0.0
        self._last_receive_time = 0.0
        self._receive_type = None

        self._can_data_lock = threading.Lock()
        self._can_id        = bytearray.fromhex("0000")
        self._can_cnc       = bytearray.fromhex("FF")
        self._can_type      = bytearray.fromhex("00")
        self._can_data      = bytearray.fromhex("00000000")
        self._can_state     = bytearray.fromhex("00")
        self._can_crc       = bytearray.fromhex("00")


    def _set_can_id(self, can_id):
        self._can_id = bytearray([can_id>>8, can_id & 0x00FF])


    def _get_can_id(self):
        return self._can_id[1] + (self._can_id[0] << 8)


    def _get_identifier(self):
        return self._identifier


    def _set_can_cnc(self, can_cnc):
        self._can_cnc[0] = can_cnc


    def _increment_can_cnc(self, can_cnc):
        if self._can_cnc[0] == 0xFE:
            self._can_cnc[0] = 0x00
        else:
            self._can_cnc[0] += 0x01


    def _reset_can_cnc(self):
        self._can_cnc[0] = 0x00


    def _unset_can_cnc(self):
        self._can_cnc[0] = 0xFF


    def _get_can_cnc(self):
        return self._can_cnc[0]


    def _set_can_type(self, can_type):
        self._can_type = can_type


    def _get_can_type(self):
        return self._can_type


    def _set_can_data(self, can_data):
        with self._can_data_lock:
            self._can_data = can_data


    def _reset_can_data(self):
        with self._can_data_lock:
            self._can_data = bytearray.fromhex("00000000")


    def _get_can_data(self):
        with self._can_data_lock:
            return self._can_data


    def _set_can_state(self, can_state):
        self._can_state = can_state


    def _get_can_state(self):
        return self._can_state


    def _calc_crc(self):
        try:
            self._can_crc = calc_crc8(self._can_id +
                                      self._can_cnc +
                                      self._can_type +
                                      self._can_data +
                                      self._can_state)
        except Exception as e:
            print(e)


    def send_once(self):
        self._send_type = self._protocol.SEND_ONCE
        self._protocol._register_to_send_list(self)


    def start_sending(self):
        if (0 < self._send_rate):
            self._send_type = self._protocol.SEND_CONSTANTLY
            self._protocol._register_to_send_list(self)
            return True
        else:
            return False


    def stop_sending(self):
        self._protocol._unregister_from_send_list(self)


    def set_send_rate(self, rate):
        if (0.0 < rate):
            self._send_rate = rate
            return True
        else:
            return False


    def get_send_rate(self):
        return self._send_rate


    def get_real_send_rate(self):
        return self._real_send_rate


    def receive_once(self):
        self._receive_type = self._protocol.RECEIVE_ONCE
        self._protocol._register_to_receive_list(self)


    def start_receiving(self):
        self._receive_type = self._protocol.RECEIVE_CONSTANTLY
        self._protocol._register_to_receive_list(self)


    def stop_receiving(self):
        self._protocol._unregister_from_receive_list(self)


    def set_receive_rate(self, rate):
        if self._protocol._configure_received_infos_rate(self, rate):
            self._receive_rate = rate

    def get_receive_rate(self):
        return self._real_receive_rate


    def get_last_receive_time(self):
        return self._last_receive_time



    def update_from_raw(self, raw, receive_time = None):

        if receive_time:
            delay = receive_time - self._last_receive_time
            if delay != 0:
                self._real_receive_rate = 1 / (delay)
                self._last_receive_time = receive_time

        self._can_cnc = raw[2]
        self._can_data = raw[4:8]
        self._can_state = raw[8]
        self._can_crc =  raw[9]


    def raw(self):
        self._calc_crc()
        return self._can_id + \
               self._can_cnc + \
               self._can_type + \
               self._can_data + \
               self._can_state + \
               self._can_crc


class OscarCmdData(OscarData):
    """
    Command data type
    """
    def __init__(self,  *args, **kwargs):
        super(OscarCmdData, self).__init__(*args, **kwargs)


class OscarInfoData(OscarData):
    """
    Info / Status data type
    """
    def __init__(self,  *args, **kwargs):
        super(OscarInfoData, self).__init__(*args, **kwargs)


class OscarAckData(OscarData):
    """
    Acknowledgment / Confirmation data type
    """
    def __init__(self,  *args, **kwargs):
        super(OscarAckData, self).__init__(*args, **kwargs)


# ----------------------- OSCAR_CAN_V2 -----------------------------------------

class OscarSteeringWheelTorqueCmdDataV2(OscarCmdData):
    """
    Steering Wheel Control by Torque Command data type
    """

    DONT_CHANGE      = 0x00
    INTERCEPTION_ON  = 0x01
    INTERCEPTION_OFF = 0x02

    LEFT_DIRECTION  = 0x02
    RIGHT_DIRECTION = 0x01

    MAX_TORQUE = 1000

    def __init__(self,  *args, **kwargs):
        super(OscarSteeringWheelTorqueCmdDataV2, self).__init__(*args, **kwargs)

        self._set_can_id(STEERING_WHEEL_TORQUE_CMD_V2)
        self._can_type  = OscarData.CAN_TYPE_CMD
        self._can_state = OscarData.CAN_STATE_WORK_AND_ACTIVE


    def _set_interception(self, interception):
        self._can_data[0] = interception


    def interception_on(self):
        self._reset_can_data()
        self._set_interception(self.INTERCEPTION_ON)
        self.send_once()


    def interception_off(self):
        self._reset_can_data()
        self._set_interception(self.INTERCEPTION_OFF)
        self.send_once()


    def _set_direction(self, direction):
        self._can_data[1] = direction


    def _set_torque(self, torque):

        if torque > 0:
            self._set_direction(self.LEFT_DIRECTION)
        else:
            self._set_direction(self.RIGHT_DIRECTION)

        torque = min(int(abs(torque * 10)), self.MAX_TORQUE)
        self._can_data[2] = torque & 0x00FF
        self._can_data[3] = torque >> 8


    def torque(self, torque):
        self._set_torque(torque)


class OscarVehicleMoveCmdDataV2(OscarCmdData):
    """
    Steering Wheel Control by Torque Command data type
    """

    DONT_CHANGE      = 0x00
    INTERCEPTION_ON  = 0x01
    INTERCEPTION_OFF = 0x02

    LEFT_DIRECTION  = 0x02
    RIGHT_DIRECTION = 0x01

    MAX_THROTTLE = 1000

    def __init__(self,  *args, **kwargs):
        super(OscarVehicleMoveCmdDataV2, self).__init__(*args, **kwargs)

        self._set_can_id(VEHICLE_ACCELERATION_CMD_V2)
        self._can_type  = OscarData.CAN_TYPE_CMD
        self._can_state = OscarData.CAN_STATE_WORK_AND_ACTIVE


    def _set_interception(self, interception):
        self._can_data[0] = interception


    def interception_on(self):
        self._reset_can_data()
        self._set_interception(self.INTERCEPTION_ON)
        self.send_once()


    def interception_off(self):
        self._reset_can_data()
        self._set_interception(self.INTERCEPTION_OFF)
        self.send_once()


    def throttle(self, throttle):
        if throttle > 0:
            throttle = int(min(throttle * 10,  self.MAX_THROTTLE))
        else:
            throttle = int(max(throttle * 10, -self.MAX_THROTTLE)) & 0xFFFF

        self._can_data[1] = throttle & 0x00FF
        self._can_data[2] = throttle >> 8


class OscarLauncherCmdDataV2(OscarCmdData):
    """
    Launcher and LED Command data type
    """

    MODE_DONT_CHANGE = 0x00
    MODE_AUTO        = 0x01
    MODE_RADIO_JOY   = 0x02
    MODE_MANUAL      = 0x03

    LED_DONT_CHANGE = 0x00
    LED_ON          = 0x01
    LED_OFF         = 0x02

    EMERGENCY_STOP_DONT_CHANGE = 0x00
    EMERGENCY_STOP_ON          = 0x01
    EMERGENCY_STOP_OFF         = 0x02

    HAND_BRAKE_DONT_CHANGE = 0x00
    HAND_BRAKE_ON          = 0x01
    HAND_BRAKE_OFF         = 0x02

    def __init__(self,  *args, **kwargs):
        super(OscarLauncherCmdDataV2, self).__init__(*args, **kwargs)

        self._set_can_id(LAUNCHER_CMD_V2)
        self._can_type  = OscarData.CAN_TYPE_CMD
        self._can_state = OscarData.CAN_STATE_WORK_AND_ACTIVE


    def _set_mode(self, mode):
        with self._can_data_lock:
            self._can_data[0] = mode


    def auto_mode(self):
        self._reset_can_data()
        self._set_mode(self.MODE_AUTO)
        self.send_once()


    def manual_mode(self):
        self._reset_can_data()
        self._set_mode(self.MODE_MANUAL)
        self.send_once()


    def _set_led(self, led):
        with self._can_data_lock:
            self._can_data[1] = led


    def led_on(self):
        self._reset_can_data()
        self._set_led(self.LED_ON)
        self.send_once()


    def led_off(self):
        self._reset_can_data()
        self._set_led(self.LED_OFF)
        self.send_once()


    def led_reverse(self):
        # got current LED state from self._protocol.launcher_info.get_led_state()
        pass


    def _set_emergency_stop(self, emergency_stop):
        with self._can_data_lock:
            self._can_data[2] = emergency_stop


    def emergency_stop_on(self):
        self._reset_can_data()
        self._set_emergency_stop(self.EMERGENCY_STOP_ON)
        self.send_once()


    def emergency_stop_off(self):
        self._reset_can_data()
        self._set_emergency_stop(self.EMERGENCY_STOP_OFF)
        self.send_once()


    def _set_hand_brake(self, hand_brake):
        with self._can_data_lock:
            self._can_data[3] = hand_brake


    def hand_brake_on(self):
        self._reset_can_data()
        self._set_hand_brake(self.HAND_BRAKE_ON)
        self.send_once()


    def hand_brake_off(self):
        self._reset_can_data()
        self._set_hand_brake(self.HAND_BRAKE_OFF)
        self.send_once()


class OscarTurnSignalsCmdDataV2(OscarCmdData):

    TURN_SIGNALS_DONT_CHANGE = 0x00
    RIGHT_TURN_SIGNAL        = 0x01
    LEFT_TURN_SIGNAL         = 0x02
    EMERGENCY_SIGNALS        = 0x03
    TURN_OFF_SIGNALS         = 0x04

    def __init__(self,  *args, **kwargs):
        super(OscarTurnSignalsCmdDataV2, self).__init__(*args, **kwargs)

        self._set_can_id(TURN_SIGNALS_CMD_V2)
        self._can_type  = OscarData.CAN_TYPE_CMD
        self._can_state = OscarData.CAN_STATE_WORK_AND_ACTIVE


    def _set_signal(self, signal):
        with self._can_data_lock:
            self._can_data[0] = signal


    def left_signal(self):
        self._reset_can_data()
        self._set_signal(self.LEFT_TURN_SIGNAL)
        self.send_once()


    def right_signal(self):
        self._reset_can_data()
        self._set_signal(self.RIGHT_TURN_SIGNAL)
        self.send_once()


    def emergency_signals(self):
        self._reset_can_data()
        self._set_signal(self.EMERGENCY_SIGNALS)
        self.send_once()


    def turn_off_signals(self):
        self._reset_can_data()
        self._set_signal(self.TURN_OFF_SIGNALS)
        self.send_once()


class OscarInfoConfigurationCmdDataV2(OscarCmdData):
    """
    Launcher and LED Command data type
    """

    DONT_CHANGE     = 0x00
    MODULE_TURN_ON  = 0x01
    MODULE_TURN_OFF = 0x02
    MODULE_RESTART  = 0x03
    SET_DELAY       = 0x04

    def __init__(self,  *args, **kwargs):
        super(OscarInfoConfigurationCmdDataV2, self).__init__(*args, **kwargs)

        self._set_can_id(INFO_CONFIGURATION_CMD_V2)
        self._can_type  = OscarData.CAN_TYPE_CMD
        self._can_state = OscarData.CAN_STATE_WORK_AND_ACTIVE


    def _set_module_cmd(self, cmd):
        with self._can_data_lock:
            self._can_data[3] = cmd


    def _set_module_can_id(self, can_id):
        with self._can_data_lock:
            self._can_data[0] = can_id[0]
            self._can_data[1] = can_id[1]


    def _set_delay(self, delay):
        with self._can_data_lock:
            delay = min(int(abs(delay) * 1000), 2500)
            if (delay > 0) and (delay < 10):
                delay = 10
            self._can_data[2] = delay/10


    def set_reply_delay(self, oscar_info_data, delay):
        self._reset_can_data()
        self._set_module_cmd(self.SET_DELAY)
        self._set_delay(delay)
        self._set_module_can_id(oscar_info_data._can_id)
        self.send_once()


    def turn_off_replies(self, oscar_info_data):
        self._reset_can_data()
        self._set_module_cmd(self.SET_DELAY)
        self._set_delay(0)
        self._set_module_can_id(oscar_info_data._can_id)
        self.send_once()


class OscarLauncherInfoDataV2(OscarInfoData):
    """
    Launcher, LED, Emergency stop and Hand brake Info / Status data type
    """

    UNKNOWN   = 0x00

    MODE_AUTO         = VEHICLE_AUTO_MODE
    MODE_MANUAL       = VEHICLE_MANUAL_MODE
    MODE_RADIO_JOY    = 0x03

    SOURCE_VEHICLE = 0x01
    SOURCE_BUTTON  = 0x02
    SOURCE_REMOTE  = 0x03
    SOURCE_AUTO    = 0x04
    SOURCE_CMD     = 0x05

    LED_ON      = 0x01
    LED_OFF     = 0x02

    EMERGENCY_STOP_ON  = 0x01
    EMERGENCY_STOP_OFF = 0x02

    HAND_BRAKE_ON  = 0x01
    HAND_BRAKE_OFF = 0x02

    def __init__(self,  *args, **kwargs):
        super(OscarLauncherInfoDataV2, self).__init__(*args, **kwargs)
        self._set_can_id(LAUNCHER_INFO_V2)
        self._can_type = OscarData.CAN_TYPE_INFO


    def get_mode(self):

        with self._can_data_lock:
            button = self._can_data[0]

        source = (button & 0b00111000) >> 3
        mode = (button & 0b00000111)

        return mode, source


    def get_led(self):
        with self._can_data_lock:
            return self._can_data[1]


    def get_emergency_stop(self):

        with self._can_data_lock:
            emergency_stop = self._can_data[2]

        source = (emergency_stop & 0b00111000) >> 3
        state = (emergency_stop & 0b00000111)

        return state, source


    def get_hand_brake(self):

        with self._can_data_lock:
            hand_brake = self._can_data[3]

        source = (hand_brake & 0b00111000) >> 3
        state = (hand_brake & 0b00000111)

        return state, source


class OscarVehicleSpeedInfoDataV2(OscarInfoData):

    def __init__(self,  *args, **kwargs):
        super(OscarVehicleSpeedInfoDataV2, self).__init__(*args, **kwargs)
        self._set_can_id(VEHICLE_SPEED_INFO_V2)
        self._can_type = OscarData.CAN_TYPE_INFO
        self._can_data = bytearray.fromhex("0000000000000000")


    def get_vehicle_speed(self):
        with self._can_data_lock:
            speed_h, speed_l = self._can_data[5], self._can_data[6]

        return ((speed_h << 8) + speed_l) / 3.6 / 100.


    def get_vehicle_encoder(self):    # 6-8 wheels turns for 256 ticks
        return self._can_data[4]      # works from some speed


    def update_from_raw(self,raw, receive_time = None):

        if receive_time:
            delay = receive_time - self._last_receive_time
            if delay != 0:
                self._real_receive_rate = 1 / (delay)
                self._last_receive_time = receive_time

        self._can_data = raw[2:10]


class OscarVehicleWheelsSpeedInfoDataV2(OscarInfoData):

    def __init__(self,  *args, **kwargs):
        super(OscarVehicleWheelsSpeedInfoDataV2, self).__init__(*args, **kwargs)
        self._set_can_id(VEHICLE_WHEELS_SPEED_INFO_V2)
        self._can_type = OscarData.CAN_TYPE_INFO
        self._can_data = bytearray.fromhex("0000000000000000")


    def get_vehicle_wheels_speed(self):
        with self._can_data_lock:
            fr_h, fr_l = self._can_data[0], self._can_data[1]
            fl_h, fl_l = self._can_data[2], self._can_data[3]
            fr_h, fr_l = self._can_data[4], self._can_data[5]
            fl_h, fl_l = self._can_data[6], self._can_data[7]

        return ((((fr_h << 8) + fr_l) * 0.01 - 67.67) / 3.6, #FR
                (((fl_h << 8) + fl_l) * 0.01 - 67.67) / 3.6, #FL
                (((fr_h << 8) + fr_l) * 0.01 - 67.67) / 3.6, #RR
                (((fl_h << 8) + fl_l) * 0.01 - 67.67) / 3.6) #RL


    def update_from_raw(self, raw, receive_time = None):

        if receive_time:
            delay = receive_time - self._last_receive_time
            if delay != 0:
                self._real_receive_rate = 1 / (delay)
                self._last_receive_time = receive_time

        self._can_data = raw[2:10]


class OscarSteeringWheelPoseVelocityInfoDataV2(OscarInfoData):

    def __init__(self,  *args, **kwargs):
        super(OscarSteeringWheelPoseVelocityInfoDataV2, self).__init__(*args, **kwargs)
        self._set_can_id(STEERING_WHEEL_POSE_VELOCITY_INFO_V2)
        self._can_type = OscarData.CAN_TYPE_INFO
        self._can_data = bytearray.fromhex("0000000000000000")


    def get_steering_wheel_angle(self):
        with self._can_data_lock:
            sw_angle_h, sw_angle_l = self._can_data[0], self._can_data[1]
        return 1.5 * twos_complement((sw_angle_h << 8) + sw_angle_l, 12)


    def get_steering_wheel_velocity(self):
        with self._can_data_lock:
            sw_velocity_h, sw_velocity_l = self._can_data[4], self._can_data[5]
        return twos_complement((sw_velocity_h << 8) + sw_velocity_l, 12)


    def get_steering_wheel_angle_and_velocity(self):

        with self._can_data_lock:
            sw_angle_h, sw_angle_l = self._can_data[0]  & 0x0F, self._can_data[1]
            sw_velocity_h, sw_velocity_l = self._can_data[4] & 0x0F, self._can_data[5]

        return (1.5 * twos_complement((sw_angle_h << 8) + sw_angle_l, 12),
                twos_complement((sw_velocity_h << 8) + sw_velocity_l, 12))


    def get_steering_fraction(self):
        return 0.1 * twos_complement(self._can_data[4] >> 4, 4)


    def update_from_raw(self, raw, receive_time = None):

        self._can_data = raw[2:10]

        if receive_time:
            delay = receive_time - self._last_receive_time
            if delay != 0:
                self._real_receive_rate = 1 / (delay)
                self._last_receive_time = receive_time


class OscarSteeringWheelEpsTorquesInfoDataV2(OscarInfoData):

    def __init__(self,  *args, **kwargs):
        super(OscarSteeringWheelEpsTorquesInfoDataV2, self).__init__(*args, **kwargs)
        self._set_can_id(STEERING_WHEEL_EPS_TORQUE_INFO_V2)
        self._can_type = OscarData.CAN_TYPE_INFO
        self._can_data = bytearray.fromhex("0000000000000000")


    def get_steering_wheel_torque(self):
        with self._can_data_lock:
            sw_torque_h, sw_torque_l = self._can_data[1], self._can_data[2]

        return twos_complement((sw_torque_h << 8) + sw_torque_l, 16) # [-32768 | 32767]


    def get_eps_torque(self):
        with self._can_data_lock:
            eps_h, eps_l = self._can_data[5], self._can_data[6]

        return 0.73 * twos_complement((eps_h << 8) + eps_l, 16) # [-20000 | 20000]


    def get_steering_wheel_and_eps_torques(self):

        with self._can_data_lock:
            sw_torque_h, sw_torque_l = self._can_data[1], self._can_data[2]
            eps_h, eps_l             = self._can_data[5], self._can_data[6]

        return (twos_complement((sw_torque_h << 8) + sw_torque_l, 16),
                0.73 * twos_complement((eps_h << 8) + eps_l, 16))


    def update_from_raw(self, raw, receive_time = None):

        if receive_time:
            delay = receive_time - self._last_receive_time
            if delay != 0:
                self._real_receive_rate = 1 / (delay)
                self._last_receive_time = receive_time

        self._can_data = raw[2:10]


# ----------------------- OSCAR_CAN_V3 -----------------------------------------
# TODO
#--------------------

# x^8 + x^2 + x^1 + x^0
def calc_crc8(data):

    crc8 = 0xFF

    for byte in data:
        crc8 ^= byte

        for i in range(8):

            if (crc8 & 0x80):
                xor_val = 0x07
            else:
                xor_val = 0x00

            crc8 = ((crc8 << 1) & 0x00FF) ^ xor_val

    return bytearray([crc8])


def twos_complement(value, bits):
    if value & (1 << (bits-1)):
        value -= 1 << bits
    return value



#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import os, sys, time

import json
from collections import OrderedDict as odict

from argparse import ArgumentParser

import matplotlib.pyplot as plt

import log


class VehicleLogViz():

    def __init__(self):
        self._path_to_vehicle_log_file = self.pars_arguments().file_path
        self.vehicle_log = self.load_vehicle_log_from_file()
        self.fig, self.ax = plt.subplots(1, 1, figsize=[22.0, 12.0], facecolor='#E0E0E0')#, axisbg='black')
        plt.subplots_adjust(bottom=0.05, right=0.98, top=0.96, left=0.04)
        # self.ax.set_aspect('equal')
        # self.ax.set_xlim([-12, 12])
        # self.ax.set_ylim([-12, 12])
        self.ax.grid(True, color='#E0E0E0')
        self.ax.set_xlabel('time, [s]')
        self.ax.set_ylabel('data')
        self.ax.patch.set_facecolor('#161616')

        plt.title(self._path_to_vehicle_log_file)

        self.sw_angle_plot, = self.ax.plot([], [],      color="#483D8B", lw=2, ls='-', label='sw_angle [ang]')
        self.sw_velocity_plot, = self.ax.plot([], [],   color="#DC143C", lw=2, ls='-', label='sw_velocity [ang/s]')
        self.sw_torque_plot, = self.ax.plot([], [],     color="#00FA9A", lw=2, ls='-', label='sw_torque')
        self.eps_torque_plot, = self.ax.plot([], [],    color="#1E90FF", lw=2, ls='--', label='eps_torque * 0.1')
        self.vehicle_speed_plot, = self.ax.plot([], [], color="#EB7A6F", lw=2, ls='-', label='vehicle_speed [cm/s]')
        self.mode_plot, = self.ax.plot([], [],          color="#FFA500", lw=1, ls='-', label='mode')

        self.legend = self.ax.legend()
        self.legend.get_frame().set_facecolor('#E0E0E0')


    def pars_arguments(self):
        parser = ArgumentParser(description="Script plots vehicle logs.")
        parser.add_argument('-n', '--name', dest='file_path', required=True, help='Full path to log file')
        return parser.parse_args()


    def load_vehicle_log_from_file(self):

        try:
            log_file_handler = open(self._path_to_vehicle_log_file, 'r')
        except Exception as err:
            log.error("Can't open file " + self._path_to_vehicle_log_file + ". " + str(err))

        try:
            vehicle_log_data = json.load(log_file_handler)
        except Exception as err:
            log.error("Can't load log from file " + self._path_to_vehicle_log_file + ". " + str(err))

        log_file_handler.close()
        return vehicle_log_data


    def plot_vehicle_log(self):

        t = []
        sw_angle = []
        sw_velocity = []
        sw_torque = []
        eps_torque = []
        vehicle_speed = []
        mode = []

        for element in self.vehicle_log["data"]:
            t.append(float(element['t']))
            sw_angle.append(float(element['sw_angle']))
            sw_velocity.append(float(element['sw_velocity']))
            sw_torque.append(float(element['sw_torque']))
            eps_torque.append(float(element['eps_torque']) * 0.1)
            vehicle_speed.append(float(element['vehicle_speed']) * 100)
            mode.append(float(element['mode']) * 100)

        self.sw_angle_plot.set_data(t, sw_angle)
        self.sw_velocity_plot.set_data(t, sw_velocity)
        self.sw_torque_plot.set_data(t, sw_torque)
        self.eps_torque_plot.set_data(t, eps_torque)
        self.vehicle_speed_plot.set_data(t, vehicle_speed)
        self.mode_plot.set_data(t, mode)

        self.ax.relim()      # make sure all the data fits
        self.ax.autoscale()

        plt.show()


if __name__ == "__main__":

    log_plotter = VehicleLogViz()
    log_plotter.plot_vehicle_log()



###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

RAW_SLP_LEN = 26
SLP_HEADER_LEN  = 8
RECEIVE_SLP_HEADER = bytearray.fromhex('AAFF0110001000C1')
SEND_SLP_HEADER    = bytearray.fromhex('AA01FF100010001C')
MAIN_SLP_BUS       = 0x1


# x^8 + x^2 + x^1 + x^0
def calc_crc8(data):

    crc8 = 0xFF

    for byte in data:
        crc8 ^= byte

        for i in range(8):

            if (crc8 & 0x80):
                xor_val = 0x07
            else:
                xor_val = 0x00

            crc8 = ((crc8 << 1) & 0x00FF) ^ xor_val
    # print("crc8: " + hex(crc8))
    return bytearray([crc8])


# x^16 + x^12 + x^5 + x^0
def calc_crc16(data):

    crc16 = 0xFFFF

    for byte in data:
        crc16 ^= byte << 8

        for i in range(8):

            if (crc16 & 0x8000):
                xor_val = 0x1021
            else:
                xor_val = 0x0000

            crc16 = ((crc16 << 1) & 0x00FFFF) ^ xor_val

    slp_crc16 = bytearray([crc16 & 0x00FF, crc16>>8])

    return slp_crc16


def create_raw_slp_from_raw_oscar_data(raw_data):
    raw_slp_data = bytearray.fromhex('0101') + \
                   bytearray([raw_data[1]]) + \
                   bytearray([raw_data[0]]) + \
                   bytearray.fromhex('00000008') + \
                   raw_data[2:]
    raw_slp_data_crc = calc_crc16(raw_slp_data)
    raw_slp = SEND_SLP_HEADER + raw_slp_data + raw_slp_data_crc
    return raw_slp


def get_slp_data_from_raw_slp(raw):
    if ((len(raw) == RAW_SLP_LEN) and (raw[0:SLP_HEADER_LEN] == RECEIVE_SLP_HEADER) and (raw[9] == MAIN_SLP_BUS)):
        return bytearray([raw[SLP_HEADER_LEN+3]]) + \
               bytearray([raw[SLP_HEADER_LEN+2]]) + \
               raw[SLP_HEADER_LEN+8:-2]
    else:
        return None


def slp_data_crc_is_ok(raw_slp_data, raw_slp_crc):
    if calc_crc16(raw_slp_data) == raw_slp_crc:
        return True
    else:
        return False


def print_raw(raw_data):
    print " ".join(hex(byte) for byte in raw_data)



#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import threading
import time

from math import cos, sin, tan, pi

import oscar_protocol
from loops import Spinner
from slp import print_raw
from oscar_control import *
from oscar_vehicle_interfaces import create_interface_by_name

import log

OSCAR_DEFAULT_CONTROL_RATE = 50
OSCAR_DEFAULT_ODOMETRY_CALC_RATE = 50
OSCAR_DEFAULT_LOG_RATE = 50
OSCAR_DEFAULT_LOG_DIR  = "/tmp/oscar_vehicle_api_logs"


class OscarVehicle(object):

    UNKNOWN                   = oscar_protocol.UNKNOWN
    VEHICLE_AUTO_MODE         = oscar_protocol.VEHICLE_AUTO_MODE
    VEHICLE_MANUAL_MODE       = oscar_protocol.VEHICLE_MANUAL_MODE

    def __init__(self, interface          = None,
                       protocol_config    = None,
                       controller         = None,
                       controller_params  = None,
                       control_loop_rate  = OSCAR_DEFAULT_CONTROL_RATE,
                       odometry_loop_rate = OSCAR_DEFAULT_ODOMETRY_CALC_RATE):

        self._vehicle_protocol  = oscar_protocol.OscarProtocol(protocol_config)
        self._vehicle_interface = create_interface_by_name(interface, self._vehicle_protocol)

        if self._vehicle_interface:
            self._vehicle_interface.start_communication()

        self.vehicle_params = {"wheel_radius":             0.37,
                               "wheel_width":              0.25,
                               "wheelbase":                2.75,
                               "axle_track":               1.64,
                               "steering_ratio":           14.8,
                               "max_steering_wheel_angle": 500,  # ~ 33.78 deg
                               "max_acceleration":         2.0,
                               "max_deceleration":         6.0}

        self._controller = create_controller_by_name(controller, self, controller_params)
        self._control_spinner = Spinner(target = self._calc_control,
                                          rate = control_loop_rate)

        self._odometry = Odometry(self)
        self._odometry_spinner = Spinner(target = self._odometry.calc_odometry,
                                           rate = odometry_loop_rate)

        self._api_vehicle_mode = self.UNKNOWN
        self._vehicle_mode_check_spinner = Spinner(target = self._check_vehicle_mode,
                                                     rate = 1)
        self._vehicle_log = None
        self._vehicle_log_dir = OSCAR_DEFAULT_LOG_DIR
        self._vehicle_log_spinner = Spinner(target = self._log_vehicle_state,
                                              rate = OSCAR_DEFAULT_LOG_RATE)

        self.start_odometry_calculation()


    def set_interface(self, interface):

        if self._vehicle_interface:
            self._stop_interface_communication()

        self._vehicle_interface = interface.create(interface)
        self._start_interface_communication()


    ### CONTROL IFACE ##########################################################

    def start_controller(self):
        if (self._controller and (not self._control_spinner.is_active())):
                self._control_spinner.start()


    def stop_controller(self):
        self._control_spinner.stop()


    def set_controller(self, controller, params = None):

        controller_is_active = self._control_spinner.is_active()

        if controller_is_active:
            self.stop_controller()

        self._controller = create_controller_by_name(controller, params)

        if self._controller and controller_is_active:
            self.start_controller()


    def set_controller_params(self, params):
        if self._controller:
            self._controller.update_params(params)


    def set_controller_rate(self, rate):
        self._control_spinner.set_rate(rate)


    def get_actual_controller_rate(self):
        return self._control_spinner.get_real_rate()


    def _calc_control(self):

        # CALL CONTROL CALC STEP HERE
        throttle, sw_torque = self._controller.calc_output()

        self._vehicle_protocol.set_vehicle_throttle(throttle)
        self._vehicle_protocol.set_steering_wheel_torque(sw_torque)


    def set_speed(self, speed = None, acceleration = None, jerk = None):
        if self._controller:
            self._controller.set_target_speed(speed        = speed,
                                              acceleration = acceleration,
                                              jerk         = jerk)


    def set_steering(self, steering_wheel_angle = None, steering_wheel_velocity = None):
        if self._controller:
            self._controller.set_target_steering(steering_wheel_angle    = steering_wheel_angle,
                                                 steering_wheel_velocity = steering_wheel_velocity)


    def set_vehicle_throttle(self, throttle):
        self._vehicle_protocol.set_vehicle_throttle(throttle)


    def set_steering_wheel_torque(self, steering_wheel_torque):
        self._vehicle_protocol.set_steering_wheel_torque(steering_wheel_torque)

    ############################################################################

    def start_odometry_calculation(self):
        if self._odometry_spinner.is_active():
            self._odometry_spinner.stop()
            self._odometry.reset()
        self._odometry_spinner.start()


    def stop_odometry_calculation(self):
        self._odometry_spinner.stop()


    def get_odometry(self):
        return self._odometry.get()


    def reset_odometry(self):
        self._odometry.reset()


    def set_odometry_calc_rate(self, rate):
        self._odometry_spinner.set_rate(rate)


    def get_actual_odometry_calc_rate(self):
        return self._odometry_spinner.get_real_rate()


    def auto_mode(self):

        if (self.get_mode() != self.VEHICLE_AUTO_MODE):

            if not self._vehicle_protocol.auto_mode():
                return False

        self._api_vehicle_mode = self.VEHICLE_AUTO_MODE
        self._vehicle_protocol.start_sending_vehicle_move_cmd()
        self._vehicle_protocol.start_sending_steering_wheel_torque_cmd()

        return True


    def manual_mode(self):

        self._vehicle_protocol.stop_sending_vehicle_move_cmd()
        self._vehicle_protocol.stop_sending_steering_wheel_torque_cmd()

        self._vehicle_protocol.set_vehicle_throttle(0)
        self._vehicle_protocol.set_steering_wheel_torque(0)

        self._api_vehicle_mode = self.VEHICLE_MANUAL_MODE

        if (self.get_mode() != self.VEHICLE_MANUAL_MODE):

            if not self._vehicle_protocol.manual_mode():
                return False

        return True


    def get_mode(self):
        mode, source = self._vehicle_protocol.get_mode()
        return mode


    def _check_vehicle_mode(self):
        mode, source = self._vehicle_protocol.get_mode()
        if self._api_vehicle_mode != mode:
            if mode == self.VEHICLE_AUTO_MODE:
                self.auto_mode()
            else:
                self.manual_mode()


    def emergency_stop(self):
        return self._vehicle_protocol.emergency_stop_on()


    def recover(self):
        return (self._vehicle_protocol.hand_brake_off() and
                self._vehicle_protocol.emergency_stop_off())


    def hand_brake(self):
        return self._vehicle_protocol.hand_brake_on()


    def led_blink(self):
        self._vehicle_protocol.led_reverse()
        for i in range(3):
            time.sleep(0.5)
            self._vehicle_protocol.led_reverse()
        return True


    def led_on(self):
        return self._vehicle_protocol.led_on()


    def led_off(self):
        return self._vehicle_protocol.led_off()


    def get_led(self):
        return self._vehicle_protocol.get_led()


    def left_turn_signal(self):
        self._vehicle_protocol.left_turn_signal()


    def right_turn_signal(self):
        self._vehicle_protocol.right_turn_signal()


    def emergency_signals(self):
        self._vehicle_protocol.emergency_signals()


    def turn_off_signals(self):
        self._vehicle_protocol.turn_off_signals()


    def get_emergency_stop(self):
        return self._vehicle_protocol.get_emergency_stop()


    def get_hand_brake(self):
        return self._vehicle_protocol.get_hand_brake()


    def get_vehicle_speed(self):
        return self._vehicle_protocol.get_vehicle_speed()


    def get_vehicle_wheels_speed(self):
        return self._vehicle_protocol.get_vehicle_wheels_speed()


    def get_steering_wheel_angle_and_velocity(self):
        return self._vehicle_protocol.get_steering_wheel_angle_and_velocity()


    def get_steering_wheel_and_eps_torques(self):
        return self._vehicle_protocol.get_steering_wheel_and_eps_torques()


    def vehicle_move_interception_on(self):
        return self._vehicle_protocol.vehicle_move_interception_on()


    def vehicle_move_interception_off(self):
        return self._vehicle_protocol.vehicle_move_interception_off()


    def start_sending_vehicle_move_cmd(self):
        return self._vehicle_protocol.start_sending_vehicle_move_cmd()


    def stop_sending_vehicle_move_cmd(self):
        return self._vehicle_protocol.stop_sending_vehicle_move_cmd()


    def steering_wheel_interception_on(self):
        return self._vehicle_protocol.steering_wheel_interception_on()


    def steering_wheel_interception_off(self):
        return self._vehicle_protocol.steering_wheel_interception_off()


    def start_sending_steering_wheel_torque_cmd(self):
        return self._vehicle_protocol.start_sending_steering_wheel_torque_cmd()


    def stop_sending_steering_wheel_torque_cmd(self):
        return self._vehicle_protocol.stop_sending_steering_wheel_torque_cmd()


    def error_report(self):
        return 'NO_ERROR'


    def _log_vehicle_state(self):
        cur_time = time.time()

        mode = self.get_mode()

        vehicle_speed         = self._vehicle_protocol.get_vehicle_speed()
        sw_angle, sw_velocity = self._vehicle_protocol.get_steering_wheel_angle_and_velocity()
        sw_torque, eps_torque = self._vehicle_protocol.get_steering_wheel_and_eps_torques()

        self._vehicle_log.add_data(cur_time,
                                   sw_angle,
                                   sw_velocity,
                                   sw_torque,
                                   eps_torque,
                                   vehicle_speed,
                                   mode)


    def start_vehicle_logger(self):
        self._vehicle_log = log.VehicleLog(self._vehicle_log_dir)
        self._vehicle_log_spinner.start()


    def change_vehicle_logger_dir(self, log_dir):
        self._vehicle_log_dir = log_dir


    def stop_vehicle_logger(self):
        self._vehicle_log_spinner.stop()
        self._vehicle_log.save()
        self._vehicle_log = None


class LexusRX450H(OscarVehicle):

    def __init__(self,  *args, **kwargs):
        super(LexusRX450H, self).__init__(*args, **kwargs)

        self.vehicle_params = {"wheel_radius":             0.37,
                               "wheel_width":              0.25,
                               "wheelbase":                2.75,
                               "axle_track":               1.64,
                               "steering_ratio":           14.8,
                               "max_steering_wheel_angle": 500,
                               "max_acceleration":         2.0,
                               "max_deceleration":         6.0}  # ~ 33.78 deg


class Odometry():

    def __init__(self, vehicle):

        self._vehicle = vehicle

        self.wheel_radius   = vehicle.vehicle_params["wheel_radius"]
        self.wheel_width    = vehicle.vehicle_params["wheel_width"]
        self.wheelbase      = vehicle.vehicle_params["wheelbase"]
        self.axle_track     = vehicle.vehicle_params["axle_track"]
        self.steering_ratio = vehicle.vehicle_params["steering_ratio"]

        self._pose_data_lock = threading.Lock()
        self.x    = 0.0
        self.y    = 0.0
        self.yaw  = 0.0
        self.dx   = 0.0
        self.dy   = 0.0
        self.dyaw = 0.0
        self.time = 0.0


    def get(self):
        return self.x, self.y, self.yaw, self.dx, self.dy, self.dyaw, self.time


    def reset(self):
        with self._pose_data_lock:
            self.x    = 0.0
            self.y    = 0.0
            self.yaw  = 0.0
            self.dx   = 0.0
            self.dy   = 0.0
            self.dyaw = 0.0
            self.time = 0.0


    def calc_odometry(self):

        cur_time = time.time()
        cur_vehicle_velocity          = self._vehicle.get_vehicle_speed()
        cur_sw_angle, cur_sw_velocity = self._vehicle.get_steering_wheel_angle_and_velocity()
        cur_vehicle_steering_angle = (cur_sw_angle * pi / 180) / self.steering_ratio

        with self._pose_data_lock:

            dt = cur_time - self.time

            self.dx = cos(self.yaw) * cur_vehicle_velocity
            self.dy = sin(self.yaw) * cur_vehicle_velocity
            self.dyaw = cur_vehicle_velocity / self.wheelbase * tan(cur_vehicle_steering_angle)

            self.x += dt * self.dx
            self.y += dt * self.dy
            self.yaw += dt * self.dyaw

            self.time = cur_time


supported_vehicle_models = {"LEXUS_RX_450H": LexusRX450H}

def create_vehicle_by_model(vehicle_model      = None,
                            interface          = None,
                            protocol_config    = None,
                            controller         = None,
                            controller_params  = None,
                            control_loop_rate  = OSCAR_DEFAULT_CONTROL_RATE,
                            odometry_loop_rate = OSCAR_DEFAULT_ODOMETRY_CALC_RATE):

    if(vehicle_model == '' or vehicle_model is None):
        log.print_warn("Vehicle model was not specified!")
        return None

    elif (vehicle_model in supported_vehicle_models):
        try:
            vehicle = supported_vehicle_models[vehicle_model](interface,
                                                              protocol_config,
                                                              controller,
                                                              controller_params,
                                                              control_loop_rate,
                                                              odometry_loop_rate)
            return vehicle
        except Exception as e:
            log.print_err("[OSCAR]: " + str(e) + "\n")
            return None

    else:
        log.print_warn("Vehicle model " + str(interface) + " is not supported!")
        return None



###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import setuptools

with open("README.md", "r") as file_with_description:
    long_description = file_with_description.read()

with open('requirements.txt') as file_with_requirements:
    requirements = file_with_requirements.read().splitlines()

setuptools.setup(
    name="oscar_vehicle_api",
    version="1.0b2",
    author="Nikolay Dema",
    author_email="ndema2301@gmail.com",
    description="OSCAR Vehicle API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords = ['oscar', 'vehicle', 'api'],
    url="https://gitlab.com/starline/oscar_vehicle_api",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.6',
)



#!/usr/bin/env python

###############################################################################
# Copyright 2020 ScPA StarLine Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import inspect
import sys
import log
import threading
import time
from math import cos, sin, tan, copysign

def available_controllers():

    controllers = {}
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            controllers.update({name: obj})

    return controllers


def create_controller_by_name(controller, vehicle, params = None):
    if controller:
        controllers = available_controllers()
        if controller in controllers:
            return controllers[controller](vehicle, params)
        else:
            log.print_warn("There is no " + str(controller) + " controller!")

    return None


class BaseController(object):

    def __init__(self, vehicle, params = None):
        self._params = {}

        self._vehicle = vehicle

        self._state_lock = threading.Lock()

        self._target_vehicle_speed        = None
        self._target_vehicle_acceleration = None
        self._target_vehicle_jerk         = None

        self._target_steering_wheel_angle    = None
        self._target_steering_wheel_velocity = None

        self._output_vehicle_throttle = None
        self._output_steering_torque  = None


    def update_params(self, params):
        self._params.update(params)


    def get_params(self):
        return self._params


    def reset():
        pass


    def set_target_speed(self, speed = None, acceleration = None, jerk = None):
        self._target_vehicle_speed        = speed
        self._target_vehicle_acceleration = acceleration
        self._target_vehicle_jerk         = jerk


    def set_target_steering(self, steering_wheel_angle = None, steering_wheel_velocity = None):
        self._target_steering_wheel_angle    = steering_wheel_angle
        self._target_steering_wheel_velocity = steering_wheel_velocity


    def calc_output():
        return self._output_vehicle_throttle, self._output_steering_wheel_torque


class PID_with_I_saturation(BaseController):

    MAX_DELAY_IN_CONTROL_LOOP = 0.05

    def __init__(self,  *args, **kwargs):
        super(BaseController, self).__init__(*args, **kwargs)

        self._last_steering_wheel_angle_error = 0.0
        self._last_control_time = 0.0

        self._integrator = 0.0

        self._params = {"steering_wheel_P": 0.1,
                        "steering_wheel_I": 0.1,
                        "steering_wheel_I_saturation": 80,
                        "steering_wheel_D": 0.01,
                        "vehicle_speed_P":  0.5}

    def reset():
        self._last_steering_wheel_angle_error = 0.0
        self._last_control_time = 0.0
        self._integrator = 0.0


    def calc_output():

        cur_time = time.time()
        dtime = cur_time - self._last_control_time

        # acceleration and throttle are the same for this controller
        if self._target_vehicle_acceleration < 0:
            self._output_vehicle_throttle = max(self._target_vehicle_acceleration, -100)
        else:
            self._output_vehicle_throttle = min(self._target_vehicle_acceleration,  100)

        if (dtime > self.MAX_DELAY_IN_CONTROL_LOOP):
            return self._output_vehicle_throttle, 0

        # steering wheel control part
        cur_vehicle_speed = self._vehicle.get_vehicle_speed()
        cur_steering_wheel_angle, cur_steering_wheel_velocity = self._vehicle.get_steering_wheel_angle_and_velocity()

        steering_wheel_angle_error = self._target_steering_wheel_angle - cur_steering_wheel_angle

        p_term = steering_wheel_angle_error * self._params['steering_wheel_P']

        integrator = self._integrator + steering_wheel_angle_error * dtime * self._params['steering_wheel_I']
        if abs(integrator) <  self._params['steering_wheel_I_saturation']:
            self._integrator = integrator

        dsteering_wheel_angle_error = self._last_steering_wheel_angle_error - steering_wheel_angle_error
        d_term = (dsteering_wheel_angle_error / dtime) * self._params['steering_wheel_D']

        self._output_steering_wheel_torque = p_term + self._integrator + d_term

        self._last_steering_wheel_angle_error = steering_wheel_angle_error
        self._last_control_time = cur_time

        return self._output_vehicle_throttle, self._output_steering_wheel_torque
